from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from fluxly.core.workflow.exceptions import NodesNotFoundException
    from fluxly.core.workflow.execution import WorkflowExecution
    from fluxly.core.workflow.graph import WorkflowGraph
    from fluxly.core.workflow.input import WorkflowInput
    from fluxly.core.workflow.metadata import WorkflowMetadata
    from fluxly.core.workflow.output import WorkflowOutput
    from fluxly.core.workflow.workflow import Workflow

__all__ = [
    "Workflow",
    "WorkflowGraph",
    "WorkflowInput",
    "WorkflowOutput",
    "WorkflowMetadata",
    "WorkflowExecution",
    "NodesNotFoundException",
]


def __getattr__(name: str):
    if name == "Workflow":
        from fluxly.core.workflow.workflow import Workflow
        return Workflow
    if name == "WorkflowGraph":
        from fluxly.core.workflow.graph import WorkflowGraph
        return WorkflowGraph
    if name == "WorkflowInput":
        from fluxly.core.workflow.input import WorkflowInput
        return WorkflowInput
    if name == "WorkflowOutput":
        from fluxly.core.workflow.output import WorkflowOutput
        return WorkflowOutput
    if name == "WorkflowMetadata":
        from fluxly.core.workflow.metadata import WorkflowMetadata
        return WorkflowMetadata
    if name == "WorkflowExecution":
        from fluxly.core.workflow.execution import WorkflowExecution
        return WorkflowExecution
    if name == "NodesNotFoundException":
        from fluxly.core.workflow.exceptions import NodesNotFoundException
        return NodesNotFoundException

    raise AttributeError(f"module 'fluxly.core.workflow' has no attribute '{name}'")



from examples.structured_demo.nodes.alpha.node import Alpha
from examples.structured_demo.nodes.beta.node import Beta
from examples.structured_demo.workflow.input import DemoWorkflowInput
from fluxly.core.workflow.workflow import Workflow


def build_demo_workflow() -> Workflow:
    workflow = Workflow(name="structured_demo", description="Structured demo workflow", version="0.1")

    alpha = Alpha(name="Alpha", description="Echoes a message", timeout_seconds=5)
    beta = Beta(name="Beta", description="Appends a suffix", timeout_seconds=5)

    workflow.add_node(alpha)
    workflow.add_node(beta)
    workflow.add_edge(alpha, beta)
    workflow.add_execution_group([beta])

    workflow.inputs = DemoWorkflowInput(
        verbose=True,
        timeout_seconds=30,
        max_retries=1,
        retry_delay_seconds=0,
        auto_generate_md=False,
    )

    return workflow



from fluxly.core.workflow import (
    NodesNotFoundException,
    Workflow,
    WorkflowExecution,
    WorkflowGraph,
    WorkflowInput,
    WorkflowMetadata,
    WorkflowOutput,
)

__all__ = [
    "Workflow",
    "WorkflowInput",
    "WorkflowOutput",
    "WorkflowMetadata",
    "WorkflowGraph",
    "WorkflowExecution",
    "NodesNotFoundException",
]



from fluxly.core.workflow.input import WorkflowInput


def build_cli_command_from_workflow_input(workflow_input: WorkflowInput, endpoint_name: str | None = None) -> str:
    flags = []
    for field_name, field in workflow_input.__pydantic_fields__.items():
        if field.json_schema_extra and field.json_schema_extra.get("exclude_from_cli"):
            continue

        value = getattr(workflow_input, field_name)
        default = field.default if field.default is not None else None

        if value is None or (isinstance(value, str) and not value.strip()) or str(value) == str(default):
            continue

        flag = f"--{field_name.replace('_', '-')}"
        if isinstance(value, bool):
            flags.append(flag if value else f"--no-{field_name.replace('_', '-')}")
        elif isinstance(value, list):
            for item in value:
                flags.append(f"{flag} {item}")
        else:
            flags.append(f"{flag} {value}")

    name = endpoint_name or "workflow"
    return f"{name} {' '.join(flags)}"



from __future__ import annotations

import uvicorn
from fastapi import FastAPI

from fluxly.core.api.handlers import (
    EndpointRunner,
    get_run_by_endpoint_handler,
    get_run_handler,
    health_handler,
)
from fluxly.core.api.models import ApiConfig
from fluxly.core.api.service import RunnerService
from fluxly.core.utils.consts import PACKAGE_VERSION
from fluxly.core.workflow.input import WorkflowInput
from fluxly.core.workflow.workflow import Workflow


def build_app(endpoints: dict[str, tuple[Workflow, type[WorkflowInput]]], config: ApiConfig) -> FastAPI:
    app = FastAPI(title="Fluxly API", version=PACKAGE_VERSION, **config.fastapi_kwargs)
    service = RunnerService()

    for endpoint_name, (workflow_template, wf_input_cls) in endpoints.items():
        runner = EndpointRunner(name=endpoint_name, workflow=workflow_template, input_cls=wf_input_cls, service=service)
        route_summary = f"Submit {endpoint_name} workflow run"
        route_description = (
            f"Workflow: {workflow_template.name}\n\n"
            f"Version: {workflow_template.version or 'N/A'}\n\n"
            f"Description: {workflow_template.description or ''}"
        )
        app.add_api_route(
            f"/{endpoint_name}/run",
            runner.submit,
            methods=["POST"],
            name=f"{endpoint_name}-run",
            summary=route_summary,
            description=route_description,
            tags=[endpoint_name],
            openapi_extra={
                "requestBody": {
                    "required": True,
                    "content": {
                        "application/json": {
                            "schema": wf_input_cls.model_json_schema()
                        }
                    },
                }
            },
        )

    app.get("/runs/{run_id}")(get_run_handler(service))
    app.get("/{endpoint}/runs/{run_id}")(get_run_by_endpoint_handler(service))
    app.get("/health")(health_handler)

    return app


def serve(endpoints: dict[str, tuple[Workflow, type[WorkflowInput]]], config: ApiConfig) -> None:
    app = build_app(endpoints, config)
    uvicorn.run(app, host=config.host, port=config.port, log_level=config.log_level, **config.uvicorn_kwargs)





from fluxly.core.exceptions import (
    APICallFailureException,
    DataErrorException,
    DataValidationFailureException,
    DependencyUnavailableException,
    InfrastructureErrorException,
    NetworkFailureException,
    PrerequisiteFailureException,
    TimeoutException,
    WorkflowException,
)

__all__ = [
    "WorkflowException",
    "TimeoutException",
    "InfrastructureErrorException",
    "DataErrorException",
    "PrerequisiteFailureException",
    "APICallFailureException",
    "NetworkFailureException",
    "DataValidationFailureException",
    "DependencyUnavailableException",
]





from typing import TYPE_CHECKING

from pydantic import BaseModel

if TYPE_CHECKING:
    from fluxly.core.workflow import Workflow

def sanitize_for_markdown(value: str) -> str:
    if isinstance(value, str):
        value = value.replace("\n", "<br>").replace("\r", "")
        value = value.replace("|", "\\|").replace("`", "\\`")
    return str(value)

def generate_markdown_table_from_dict(title: str, data_dict: dict) -> str:
    table = f"## {title}\n"
    table += "| Attribute   | Description                                   |\n"
    table += "|-------------|-----------------------------------------------|\n"

    for key, value in data_dict.items():
        table += f"| **{key}**  | {value} |\n"

    return table

def generate_markdown_table_from_basemodel(title: str, model: BaseModel | None) -> str:
    if model is None:
        return generate_markdown_table_from_dict(title, {"Info": "No data"})
    rows = []
    schema = model.model_json_schema()

    for parameter_name, parameter_configuration in schema.get('properties', {}).items():
        if parameter_configuration.get('exclude_from_documentation', False):
            continue

        formatted_parameter_name = parameter_name.replace('_', '-') if parameter_configuration.get('convert_underscores', True) else parameter_name
        parameters_required = schema.get('required', [])
        parameter_types = [parameter_configuration.get('type')] if 'type' in parameter_configuration else [
            param.get('type') for param in parameter_configuration.get('anyOf', [])
        ]
        model_dict = model.model_dump()
        field_value = model_dict.get(parameter_name, 'None')

        field_value = sanitize_for_markdown(str(field_value))

        additional_tags = []
        if parameter_configuration.get('convert_underscores', False):
            additional_tags.append("convert_underscores")
        if parameter_configuration.get('exclude_from_cli', False):
            additional_tags.append("exclude_from_cli")

        row = {
            'name': formatted_parameter_name,
            'description': parameter_configuration.get('description', 'None'),
            'default': parameter_configuration.get('default', 'None'),
            'required': "True" if parameter_name in parameters_required else "False",
            'type': ", ".join(filter(None, parameter_types)) if parameter_types else "Unknown",
            'value': field_value,  # Adding actual field value
            'additional_tags': ", ".join(additional_tags) if additional_tags else "None",
        }
        rows.append(row)

    table = f"## {title}\n"
    table += "| Attribute | Description | Default | Required | Type | Value | Additional Tags |\n"
    table += "|-----------|-------------|---------|----------|------|-------|----------------|\n"

    for row in rows:
        table += (f"| **{row['name']}** "
                  f"| {row['description']} "
                  f"| {row['default']} "
                  f"| {row['required']} "
                  f"| {row['type']} "
                  f"| {row['value']} "
                  f"| {row['additional_tags']} "
                  f"|\n")

    return table

def generate_graph_diagram(workflow: "Workflow", output_filename: str) -> None:
    # Resolve filename and output format (the diagrams library expects filename without extension)
    import importlib
    import os
    try:
        diagrams_mod = importlib.import_module("diagrams")
        languages_mod = importlib.import_module("diagrams.programming.language")
        Diagram = diagrams_mod.Diagram
        DiagramsEdge = diagrams_mod.Edge
        Python = languages_mod.Python
    except Exception:
        print("[DocsGenerator] Skipping diagram generation: diagrams library not available. Install it e.g.: pip install diagrams && brew install graphviz")
        return

    base_name = os.path.basename(output_filename)
    if "." in base_name:
        file_root, file_ext = os.path.splitext(output_filename)
        diagram_filename = file_root
        outformat = file_ext.lstrip(".") or "png"
    else:
        diagram_filename = output_filename
        outformat = "png"

    graph = workflow._graph

    try:
        with Diagram(f"{workflow.name} Nodes", filename=diagram_filename, outformat=outformat, show=False):
            diagram_nodes: dict[str, Python] = {}

            # Build node palette from workflow graph
            nodes_collection = workflow.get_nodes()
            node_names = [n.name for n in nodes_collection]

            for node_name in node_names:
                if node_name not in diagram_nodes:
                    diagram_nodes[node_name] = Python(node_name)

            # Draw edges from the graph
            for e in graph.edges:
                src_name = e.source
                dst_name = e.destination

                src_node = diagram_nodes.get(src_name) or Python(src_name)
                diagram_nodes[src_name] = src_node

                dst_node = diagram_nodes.get(dst_name) or Python(dst_name)
                diagram_nodes[dst_name] = dst_node

                if e.condition is not None:
                    src_node >> DiagramsEdge(label="Next Step with Condition", color="red", style="dashed") >> dst_node
                else:
                    src_node >> DiagramsEdge(label="Next Step", color="black") >> dst_node
    except Exception:
        print("[DocsGenerator] Skipping diagram generation: Graphviz not found. Install it e.g.: brew install graphviz")




import threading
import time
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Annotated
from uuid import uuid4

from pydantic import BaseModel, Field, PrivateAttr

from fluxly.core.exceptions import TimeoutException, WorkflowException
from fluxly.core.node.error import NodeError
from fluxly.core.node.execution import NodeExecution
from fluxly.core.status import StatusCodes
from fluxly.core.workflow.input import WorkflowInput
from fluxly.core.workflow.metadata import WorkflowMetadata
from fluxly.services import LoggerConfig, LoggerService


class Node(ABC, BaseModel):
    name: Annotated[str, Field(..., max_length=30, min_length=3, description="The name of the node.")]
    description: Annotated[str, Field(max_length=128, min_length=3, description="The description of the node.")] | None = None
    timeout_seconds: Annotated[int, Field(gt=0, description="Timeout for the node in seconds.")] | None = None
    max_retries: Annotated[int, Field(ge=0, description="Maximum number of run attempts allowed in case of failure.")] = 0
    retry_delay_seconds: Annotated[int, Field(ge=0, description="Delay between retries in seconds.")] = 0

    _id: str = PrivateAttr(default_factory=lambda: str(uuid4()))
    _executions: list[NodeExecution] = PrivateAttr(default_factory=list)
    _workflow_input: WorkflowInput | None = PrivateAttr(default=None)
    _workflow_metadata: WorkflowMetadata | None = PrivateAttr(default=None)
    _logger: LoggerService = LoggerService(config=LoggerConfig())

    @property
    def id(self) -> str:
        return self._id

    @property
    def executions(self) -> list[NodeExecution]:
        return self._executions

    @property
    def workflow_input(self) -> WorkflowInput | None:
        return self._workflow_input

    @property
    def workflow_metadata(self) -> WorkflowMetadata | None:
        return self._workflow_metadata

    def _set_workflow_context(self, workflow_input: WorkflowInput, workflow_metadata: WorkflowMetadata) -> None:
        self._workflow_input = workflow_input
        self._workflow_metadata = workflow_metadata

    @abstractmethod
    def _logic(self) -> None:
        raise NotImplementedError()

    @property
    def current_execution(self) -> NodeExecution:
        return self._executions[-1]

    @property
    def last_execution(self) -> NodeExecution:
        return self._executions[-1]

    @property
    def attempt(self) -> int:
        return len(self._executions)

    def execute(self) -> None:
        while self.attempt <= self.max_retries:
            try:
                self._start_node_execution()
                self.on_start()

                self._run_with_timeout()
                self.on_success()
                break
            except Exception as e:
                self.on_failure(e)
                if not self._handle_retry(e):
                    raise e
            finally:
                self._finalize_node_execution()
                self.on_finish()

    def _create_execution(self) -> NodeExecution:
        return NodeExecution()

    def _start_node_execution(self) -> None:
        execution = self._create_execution()
        execution._attempt = self.attempt + 1
        execution.metadata.start_time = datetime.now()
        execution.status = StatusCodes.IN_PROGRESS
        self._executions.append(execution)

    def _run_with_timeout(self) -> None:
        result: list[Exception | None] = [None]

        def runner() -> None:
            try:
                self._logic()
            except Exception as e:
                result[0] = e

        thread = threading.Thread(target=runner, daemon=True)
        thread.start()
        thread.join(timeout=self.timeout_seconds)

        if thread.is_alive():
            self._handle_exception(TimeoutException())

        if result[0] is not None:
            self._handle_exception(result[0])

    def _handle_retry(self, error: Exception) -> bool:
        if self.attempt >= self.max_retries:
            self._logger.error(f"{self.name} failed: {error}. Retries exhausted.")
            return False

        self._logger.warning(f"{self.name} failed: {error}. Retrying in {self.retry_delay_seconds}s...")
        time.sleep(self.retry_delay_seconds)
        return True

    def _finalize_node_execution(self) -> None:
        current = self.current_execution

        if current.metadata.end_time is None:
            current.metadata.end_time = datetime.now()

        if current.status == StatusCodes.IN_PROGRESS:
            current.status = StatusCodes.COMPLETED

    def _handle_exception(self, exception: Exception | WorkflowException | None) -> None:
        if exception is None:
            return

        current = self.current_execution
        status = exception.exit_code if isinstance(exception, WorkflowException) else StatusCodes.FAILED
        current.status = status
        current.error = NodeError(
            status=status,
            exception_class_name=exception.__class__.__name__,
            exception_message=str(exception),
        )
        raise exception

    def on_start(self) -> None:
        """Called when a node execution starts."""
        pass

    def on_success(self) -> None:
        """Called when a node execution completes successfully."""
        pass

    def on_failure(self, error: Exception) -> None:
        """Called when a node execution fails."""
        pass

    def on_finish(self) -> None:
        """Called at the end of any node execution (success or failure)."""
        pass


    def __hash__(self) -> int:
        return hash(self.name)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Node):
            return NotImplemented

        if other is self:
            return True



import logging
import sys
from typing import Any

from loguru import logger
from pydantic import BaseModel, ConfigDict

from fluxly.services.logger.config import LoggerConfig


class LoggerService(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    config: LoggerConfig
    handlers: list[logging.Handler] = []
    extra_params: dict[str, str] = {}

    def model_post_init(self, __context: Any) -> None:
        logger.remove()
        self.set_extra_params(extra_params={})

    def set_extra_params(self, extra_params: dict[str, str]) -> None:
        self.extra_params = extra_params
        self.configure()

    def configure(self) -> None:
        log_handler: dict[str, Any] = {
            "sink": sys.stdout,
            "level": self.config.level,
            "format": self.__format_record,
            "diagnose": False,
        }
        config: dict[str, Any] = {
            "handlers": [log_handler],
            "extra": self.extra_params,
        }
        if self.config.file:
            file_handler = dict(log_handler)
            file_handler["sink"] = self.config.file
            file_handler["serialize"] = True
            config["handlers"].append(file_handler)

        for handler in self.handlers:
            custom_handler = dict(log_handler)
            custom_handler["sink"] = handler
            config["handlers"].append(custom_handler)

        logger.configure(**config)  # type: ignore[arg-type]

    def add_handler(self, handler: logging.Handler) -> None:
        self.handlers.append(handler)
        self.configure()

    def remove_handler(self, handler: logging.Handler) -> None:
        if handler in self.handlers:
            self.handlers.remove(handler)
        self.configure()

    def info(self, msg: str | Any) -> None:
        logger.info(self.__wrap_message(msg))

    def debug(self, msg: str | Any) -> None:
        logger.debug(self.__wrap_message(msg))

    def warning(self, msg: str | Any) -> None:
        logger.warning(self.__wrap_message(msg))

    def error(self, msg: str | Any) -> None:
        logger.error(self.__wrap_message(msg))

    def exception(self, msg: str | Any) -> None:
        logger.exception(self.__wrap_message(msg))

    def __wrap_message(self, msg: str | Any) -> str:
        if not isinstance(msg, str):
            msg = str(msg)
        return msg if len(msg) <= self.config.max_print_length else msg[: self.config.max_print_length - 3] + "..."

    def __format_record(self, record: dict[str, Any]) -> str:
        dt_local = record["time"].astimezone()  # keep local timezone
        time_local_str = dt_local.strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]
        epoch_ms = int(record["time"].timestamp() * 1000)
        level_name = record["level"].name

        extra_keys = list(self.extra_params.keys())
        extras_section = ""
        if extra_keys:
            extras_values = " ".join(
                f"{key}={record['extra'].get(key, '')}" for key in extra_keys
            )
            extras_section = f"- {extras_values} "

        prefix = f"{time_local_str} {epoch_ms} - {level_name} {extras_section}- "

        message = record["message"]
        if not isinstance(message, str):
            message = str(message)

        # escape braces for loguru
        message = message.replace("{", "{{").replace("}", "}}")

        lines = message.splitlines() or [""]
        return "\n".join(prefix + line for line in lines) + '\n'



from datetime import datetime, timedelta


def format_datetime_to_humanreadable(dt: datetime | None) -> str | None:
    if not dt:
        return None
    return dt.strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]


def format_timedelta_to_humanreadable(td: timedelta | None) -> str | None:
    if not td:
        return None
    total_seconds = td.total_seconds()
    hours, remainder = divmod(total_seconds, 3600)
    minutes, seconds = divmod(remainder, 60)
    milliseconds = int((seconds - int(seconds)) * 1000)
    return f"{int(hours)}h {int(minutes)}m {int(seconds)}s {milliseconds}ms"



from fluxly.core.node import (
    Node,
    NodeError,
    NodeExecution,
    NodeMetadata,
    NodeOutput,
)

__all__ = [
    "Node",
    "NodeMetadata",
    "NodeOutput",
    "NodeError",
    "NodeExecution",
]



from typing import Annotated
from uuid import uuid4

from pydantic import BaseModel, Field, PrivateAttr, computed_field

from fluxly.core.node.error import NodeError
from fluxly.core.node.metadata import NodeMetadata
from fluxly.core.node.output import NodeOutput
from fluxly.core.status import StatusCodes


class NodeExecution(BaseModel):
    metadata: Annotated[NodeMetadata, Field(description="Per-attempt execution metadata")] = NodeMetadata()
    status: Annotated[StatusCodes, Field(description="Node execution status")] = StatusCodes.WAITING
    output: Annotated[NodeOutput, Field(description="Per-attempt output state")] = NodeOutput()
    error: Annotated[NodeError | None, Field(description="Error for this attempt if failed")] = None

    _id: str = PrivateAttr(default_factory=lambda: str(uuid4()))
    _attempt: int = PrivateAttr(default=1)

    @computed_field
    @property
    def id(self) -> str:
        return self._id

    @computed_field
    @property
    def attempt(self) -> int:
        return self._attempt

    def __str__(self) -> str:
        return self.model_dump_json(indent=2)

    def __repr__(self) -> str:
        return self.__str__()



import sys

from pydantic import BaseModel, PrivateAttr

from fluxly.core.api.server import ApiConfig, serve
from fluxly.core.cli.generator import build_click_group_with_commands
from fluxly.core.workflow.input import WorkflowInput
from fluxly.core.workflow.workflow import Workflow


class Fluxly(BaseModel):
    # TODO: debug: bool = False
    # TODO: max_workers: int | None = None
    _endpoints: dict[str, tuple[Workflow, type[WorkflowInput]]] = PrivateAttr(default_factory=dict)
    _api_config: ApiConfig = PrivateAttr(default_factory=ApiConfig)

    def add_endpoint(self, endpoint_name: str, workflow: Workflow, workflow_input_cls: type[WorkflowInput]) -> None:
        if endpoint_name in self._endpoints:
            raise ValueError(f"Endpoint {endpoint_name} already registered")
        if not isinstance(workflow, Workflow):
            raise TypeError("workflow must be a Workflow instance (pre-built with nodes and edges)")
        self._endpoints[endpoint_name] = (workflow, workflow_input_cls)

    def add_command(self, endpoint_name: str, workflow: Workflow, workflow_input_cls: type[WorkflowInput]) -> None:
        self.add_endpoint(endpoint_name, workflow, workflow_input_cls)

    def configure_api(self, config: ApiConfig) -> None:
        self._api_config = config

    def run_api(self) -> None:
        serve(self._endpoints, self._api_config)

    def run_cli(self) -> None:
        click_group = build_click_group_with_commands(self._endpoints)
        click_group()

    def run(self) -> None:
        if len(sys.argv) > 1:
            self.run_cli()
        else:
            self.run_api()




from typing import TYPE_CHECKING

from fluxly.core.docs_generator.utils import (
    generate_graph_diagram,
    generate_markdown_table_from_basemodel,
    generate_markdown_table_from_dict,
)

if TYPE_CHECKING:
    from fluxly.core.workflow import Workflow

def generate_workflow_documentation(workflow: "Workflow",
                                  md_output_path: str = "workflow_documentation.md",
                                  diagram_output_path: str = "workflow_diagram.png") -> None:
    amount_of_nodes = len(workflow.get_nodes())
    amount_of_execution_groups = len(workflow.get_execution_groups())

    workflow_config = {
        "Name": workflow.name,
        "Description": workflow.description,
        "Version": workflow.version,
        "Total Nodes Count": amount_of_nodes,
        "Execution Groups Count": amount_of_execution_groups,
        "Endpoint Type": workflow.endpoint_type,
        "Endpoint Name": workflow.endpoint_name,
    }

    workflow_config_table = generate_markdown_table_from_dict('⚙️ Workflow Configuration', workflow_config)
    workflow_input_table = generate_markdown_table_from_basemodel('📝 Workflow Input', workflow.inputs)

    execution_tables: list[str] = []
    if workflow.attempt > 0:
        current = workflow.current_execution
        workflow_metadata_table = generate_markdown_table_from_basemodel('📊 Workflow Metadata', current.metadata)
        workflow_output_table = generate_markdown_table_from_basemodel('📤 Workflow Output', current.output)
        execution_tables.extend([workflow_metadata_table, workflow_output_table])
    else:
        execution_tables.append(
            generate_markdown_table_from_dict('📊 Workflow Executions', {"Info": "No executions yet"})
        )

    generate_graph_diagram(workflow=workflow, output_filename=diagram_output_path)

    node_tables = []
    for idx, node in enumerate(workflow.get_nodes(), start=1):
        node_config = {
            "Name": node.name,
            "Description": node.description,
            "Timeout (seconds)": node.timeout_seconds,
            "Max Retries": node.max_retries,
            "Retry delay (seconds)": node.retry_delay_seconds,
        }

        node_config_table = generate_markdown_table_from_dict(f'⚙️ Step {idx}# - Configuration', node_config)

        node_tables.append(f"# 👾 Node {idx}# - {node.name}")
        node_tables.append(node_config_table)

    nodes_markdown = "\n\n".join(node_tables)

    markdown_content = (
        f"# 🕹️ Workflow:\n\n"
        f"{workflow_config_table}\n\n"
        f"{workflow_input_table}\n\n"
        f"{''.join(execution_tables)}\n\n"
        f"\n\n\n\n# 👾 Nodes:\n\n"
        f"\n\n## 🖼️ Nodes Diagram\n\n![Nodes Diagram]({diagram_output_path})\n\n"
        f"{nodes_markdown}\n\n"

    )

    with open(md_output_path, "w", encoding="utf-8") as f:
        f.write(markdown_content)

    print(f"Markdown file '{md_output_path}' generated successfully!")



from typing import Annotated

from pydantic import Field

from fluxly.core.node import NodeOutput


class AlphaNodeOutput(NodeOutput):
    echoed: Annotated[str, Field(description="Echoed message from Alpha")] = ""



from typing import Annotated
from uuid import uuid4

from pydantic import BaseModel, Field, PrivateAttr, computed_field

from fluxly.core.status import StatusCodes
from fluxly.core.workflow.metadata import WorkflowMetadata
from fluxly.core.workflow.output import WorkflowOutput


class WorkflowExecution(BaseModel):
    metadata: Annotated[WorkflowMetadata, Field(description="Per-attempt workflow execution metadata")] = WorkflowMetadata()
    status: Annotated[StatusCodes, Field(description="Per-attempt workflow execution status")] = StatusCodes.WAITING
    output: Annotated[WorkflowOutput, Field(description="Per-attempt workflow output state")] = WorkflowOutput()

    _id: str = PrivateAttr(default_factory=lambda: str(uuid4()))
    _attempt: int = PrivateAttr(default=1)

    @computed_field
    @property
    def attempt(self) -> int:
        return self._attempt

    @computed_field
    @property
    def id(self) -> str:
        return self._id

    def __str__(self) -> str:
        return self.model_dump_json(indent=2)

    def __repr__(self) -> str:
        return self.__str__()



from typing import Annotated

from pydantic import BaseModel, Field, computed_field

from fluxly.core.utils.types import DatetimeReadable, TimedeltaReadable


class WorkflowMetadata(BaseModel):
    start_time: Annotated[DatetimeReadable, Field(description="Start datetime of the workflow execution.")] = None
    end_time: Annotated[DatetimeReadable, Field(description="End datetime of the workflow execution.")] = None

    @computed_field
    @property
    def process_time(self) -> TimedeltaReadable:
        if self.start_time and self.end_time:
            return self.end_time - self.start_time
        return None

    def __str__(self) -> str:
        return self.model_dump_json(indent=2)



from typing import Annotated

from pydantic import BaseModel, Field


class WorkflowInput(BaseModel):
    model_config = {"extra": "forbid"}
    verbose: Annotated[bool, Field(description="Print more details for debug")] = True
    timeout_seconds: Annotated[int | None, Field(default=None, gt=0, description="Timeout for the workflow in seconds.")] = None
    max_retries: Annotated[int, Field(ge=0, description="Maximum number of run attempts allowed in case of failure.")] = 0
    retry_delay_seconds: Annotated[int, Field(ge=0, description="Delay between retries in seconds.")] = 0
    auto_generate_md: Annotated[bool, Field(description="Automatically generate markdown file for the workflow input documentation")] = False
    md_file_path: Annotated[str, Field(description="Path to save the generated markdown file.")] = "workflow_documentation.md"
    diagram_file_path: Annotated[str, Field(description="Path to save the generated workflow graph diagram image (png).")] = "workflow_diagram.png"
    # Annotated[int, Field(description="Something.",
    # json_schema_extra={
    #   'exclude_from_cli': 'True',
    #   'convert_underscores': False,
    #   'exclude_from_documentation': False
    # })]



#!/usr/bin/env python3
import argparse
import re


def compute_next_version(current: str, bump: str) -> str:
    match = re.match(r"^(\d+)\.(\d+)\.(\d+)", current)
    if match:
        major, minor, patch = map(int, match.groups())
    else:
        # fallback: try X.Y
        match = re.match(r"^(\d+)\.(\d+)", current)
        if not match:
            raise SystemExit(f"Could not parse version from: {current!r}")
        major, minor = map(int, match.groups())
        patch = 0

    if bump == "major":
        major += 1
        minor = 0
        patch = 0
    elif bump == "minor":
        minor += 1
        patch = 0
    else:
        patch += 1

    return f"{major}.{minor}.{patch}"


def main() -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument("--current", required=True)
    parser.add_argument("--bump", choices=["major", "minor", "patch"], default="patch")
    args = parser.parse_args()

    next_version = compute_next_version(args.current, args.bump)
    print(next_version)
    return 0


if __name__ == "__main__":
    raise SystemExit(main())




import time
import unittest

from fastapi import status
from fastapi.testclient import TestClient

from fluxly.core.api.server import ApiConfig, build_app
from fluxly.core.node.node import Node
from fluxly.core.status import StatusCodes
from fluxly.core.workflow.input import WorkflowInput
from fluxly.core.workflow.workflow import Workflow


class _NoopNode(Node):
    def _logic(self) -> None:
        return None


def _build_simple_workflow() -> Workflow:
    wf = Workflow(
        name="ApiTestWF",
        description="API test workflow",
        version="0.0.1",
        inputs=WorkflowInput(verbose=False, timeout_seconds=5),
    )

    wf.add_node(_NoopNode(name="noop-node", description="No operation"))
    return wf


class TestApiServer(unittest.TestCase):
    def setUp(self) -> None:
        wf = _build_simple_workflow()
        endpoints = {"test": (wf, WorkflowInput)}
        app = build_app(endpoints=endpoints, config=ApiConfig())
        self.client = TestClient(app)

    def test_health_endpoint(self) -> None:
        res = self.client.get("/health")
        self.assertEqual(res.status_code, status.HTTP_200_OK)
        self.assertEqual(res.json(), {"status": "ok"})

    def test_submit_run_and_fetch_status(self) -> None:
        res = self.client.post("/test/run", json={})
        self.assertEqual(res.status_code, status.HTTP_202_ACCEPTED)
        data = res.json()
        self.assertIn("run_id", data)
        self.assertEqual(data["endpoint"], "test")
        self.assertIn(data["status"], [s.name for s in StatusCodes])

        run_id = data["run_id"]

        # Poll for completion (or failure) briefly
        non_terminal = {StatusCodes.WAITING.name, StatusCodes.IN_PROGRESS.name}
        terminal_statuses = {s.name for s in StatusCodes} - non_terminal
        status_value = data["status"]
        deadline = time.time() + 2
        while status_value not in terminal_statuses and time.time() < deadline:
            time.sleep(0.01)
            res2 = self.client.get(f"/runs/{run_id}")
            self.assertEqual(res2.status_code, status.HTTP_200_OK)
            status_value = res2.json()["status"]

        # Verify final record shape (allow any valid status to avoid flakiness)
        res3 = self.client.get(f"/runs/{run_id}")
        self.assertEqual(res3.status_code, status.HTTP_200_OK)
        final = res3.json()
        self.assertEqual(final["run_id"], run_id)
        self.assertEqual(final["endpoint"], "test")
        self.assertIn(final["status"], {s.name for s in StatusCodes})

        res4 = self.client.get(f"/test/runs/{run_id}")
        self.assertEqual(res4.status_code, status.HTTP_200_OK)

        res5 = self.client.get(f"/wrong/runs/{run_id}")
        self.assertEqual(res5.status_code, status.HTTP_404_NOT_FOUND)

    def test_submit_validation_error(self) -> None:
        # retry_delay_seconds must be >= 0; negative should 422
        res = self.client.post("/test/run", json={"retry_delay_seconds": -1})
        self.assertEqual(res.status_code, status.HTTP_422_UNPROCESSABLE_CONTENT)

    def test_get_nonexistent_run_returns_404(self) -> None:
        res = self.client.get("/runs/does-not-exist")
        self.assertEqual(res.status_code, status.HTTP_404_NOT_FOUND)
        res2 = self.client.get("/test/runs/does-not-exist")
        self.assertEqual(res2.status_code, status.HTTP_404_NOT_FOUND)


if __name__ == "__main__":
    unittest.main()





from __future__ import annotations

import copy
import threading
from datetime import UTC, datetime
from typing import Any
from uuid import uuid4

from fluxly.core.api.models import RunRecord
from fluxly.core.status import StatusCodes
from fluxly.core.workflow.input import WorkflowInput
from fluxly.core.workflow.models import EndpointType
from fluxly.core.workflow.workflow import Workflow


class RunnerService:
    def __init__(self) -> None:
        self._runs: dict[str, RunRecord] = {}

    def submit(self, endpoint: str, workflow: Workflow, input_cls: type[WorkflowInput], values: dict[str, Any]) -> RunRecord:
        wf = copy.deepcopy(workflow)
        wf.inputs = input_cls(**values)

        run_id = wf.run_id or str(uuid4())
        wf.assign_run_id(run_id)
        wf.assign_trigger(endpoint_type=EndpointType.API, endpoint_name=endpoint)
        record = RunRecord(
            run_id=run_id,
            endpoint=endpoint,
            workflow_name=wf.name,
            workflow_version=wf.version,
            submitted_at=datetime.now(UTC).isoformat(),
            status=StatusCodes.WAITING.name,
        )
        self._runs[run_id] = record

        def _run() -> None:
            rec = self._runs[run_id]
            rec.status = StatusCodes.IN_PROGRESS.name
            rec.started_at = datetime.now(UTC).isoformat()
            rec.workflow_id = wf.id
            try:
                wf.execute()
                latest = wf.last_execution
                rec.status = latest.status.name
                rec.executions = wf.executions
            except Exception as e:
                latest = wf.last_execution
                rec.status = latest.status.name if latest else StatusCodes.FAILED.name
                rec.executions = wf.executions
                rec.error = str(e)

        threading.Thread(target=_run, daemon=True).start()
        return record

    def get(self, run_id: str) -> RunRecord | None:
        return self._runs.get(run_id)






from examples.structured_demo.nodes.beta.execution import BetaNodeExecution
from examples.structured_demo.workflow.input import DemoWorkflowInput
from fluxly.node import Node


class Beta(Node):
    _workflow_input: DemoWorkflowInput

    @property
    def current_execution(self) -> BetaNodeExecution:
        return super().current_execution

    @property
    def last_execution(self) -> BetaNodeExecution:
        return super().last_execution

    @property
    def workflow_input(self) -> DemoWorkflowInput:
        return self._workflow_input

    def _create_execution(self) -> BetaNodeExecution:
        return BetaNodeExecution()

    def _logic(self) -> None:
        suffix = self.workflow_input.suffix
        if not suffix:
            suffix = "!"

        self.current_execution.output.combined = f"processed{suffix}"
        self._logger.info(f"Beta produced: {self.current_execution.output.combined}")



import os
import tempfile
import unittest

from fluxly.core.docs_generator.generator import generate_workflow_documentation
from fluxly.core.node.node import Node
from fluxly.core.workflow.input import WorkflowInput
from fluxly.core.workflow.workflow import Workflow


class _NoopNode(Node):
    def _logic(self) -> None:
        return None


def _build_simple_workflow() -> Workflow:
    wf = Workflow(
        name="DocsGenWF",
        description="Docs generator test",
        version="0.1.0",
        inputs=WorkflowInput(auto_generate_md=False),
    )

    n1 = _NoopNode(name="A-test", description="First")
    n2 = _NoopNode(name="B-test", description="Second")

    wf.add_node(n1)
    wf.add_node(n2)
    wf.add_edge(n1, n2)

    return wf


class TestDocsGenerator(unittest.TestCase):
    def test_generate_workflow_documentation_creates_markdown_and_references_sections(self) -> None:
        wf = _build_simple_workflow()

        with tempfile.TemporaryDirectory() as tmpdir:
            md_path = os.path.join(tmpdir, "wf.md")
            diagram_path = os.path.join(tmpdir, "wf.png")

            generate_workflow_documentation(
                workflow=wf,
                md_output_path=md_path,
                diagram_output_path=diagram_path,
            )

            self.assertTrue(os.path.exists(md_path))

            with open(md_path, encoding="utf-8") as f:
                content = f.read()

            self.assertIn("# 🕹️ Workflow:", content)
            self.assertIn("⚙️ Workflow Configuration", content)
            self.assertIn("📝 Workflow Input", content)
            self.assertIn("# 👾 Nodes:", content)
            self.assertIn("## 🖼️ Nodes Diagram", content)
            self.assertIn("Node 1# - A-test", content)
            self.assertIn("Node 2# - B-test", content)





from enum import Enum


class StatusCodes(Enum):
    UNKNOWN = -1
    COMPLETED = 0
    FAILED = 1

    WAITING = 21
    IN_PROGRESS = 22

    NO_VERDICT = 41
    PARTIAL_VERDICT = 42

    TIMED_OUT = 61
    CANCELLED = 62
    INFRASTRUCTURE_ERROR = 63
    DATA_ERROR = 64

    PREREQUISITE_FAIL = 81
    API_CALL_FAILURE = 82
    NETWORK_FAILURE = 83

    DATA_VALIDATION_FAILURE = 84
    DEPENDENCY_UNAVAILABLE = 85



from enum import Enum
from typing import Annotated

from pydantic import BaseModel, Field

from fluxly.core.status import StatusCodes


class NodeError(BaseModel):
    status: Annotated[Enum, Field(description="Node error status")] = StatusCodes.FAILED
    exception_class_name: Annotated[str | None, Field(description="Exception class name")] = None
    exception_message: Annotated[str | None, Field(description="Error message")] = None

    def __str__(self) -> str:
        return self.model_dump_json(indent=2)

    def __repr__(self) -> str:
        return self.__str__()



from enum import Enum

from fluxly.core.status import StatusCodes


class WorkflowException(Exception):
    exit_code: Enum | int | None = None

    def __init__(self, message: str = ""):
        if not isinstance(self.exit_code, Enum):
            raise TypeError(f"WorkflowException exit_code must be an instance of Enum, Actual: {type(self.exit_code).__name__}")

        if not (0 <= self.exit_code.value <= 255):
            raise ValueError(f"WorkflowException exit_code value must be between 0 and 255, Actual: {self.exit_code.value}")

        super().__init__(message)

class TimeoutException(WorkflowException):
    exit_code: Enum = StatusCodes.TIMED_OUT

class InfrastructureErrorException(WorkflowException):
    exit_code: Enum = StatusCodes.INFRASTRUCTURE_ERROR

class DataErrorException(WorkflowException):
    exit_code: Enum = StatusCodes.DATA_ERROR

class PrerequisiteFailureException(WorkflowException):
    exit_code: Enum = StatusCodes.PREREQUISITE_FAIL

class APICallFailureException(WorkflowException):
    exit_code: Enum = StatusCodes.API_CALL_FAILURE

class NetworkFailureException(WorkflowException):
    exit_code: Enum = StatusCodes.NETWORK_FAILURE

class DataValidationFailureException(WorkflowException):
    exit_code: Enum = StatusCodes.DATA_VALIDATION_FAILURE

class DependencyUnavailableException(WorkflowException):
    exit_code: Enum = StatusCodes.DEPENDENCY_UNAVAILABLE



from __future__ import annotations

from typing import Any

from pydantic import BaseModel

from fluxly.core.workflow.execution import WorkflowExecution


class ApiConfig(BaseModel):
    host: str = "127.0.0.1"
    port: int = 8000
    log_level: str = "info"
    fastapi_kwargs: dict[str, Any] = {}
    uvicorn_kwargs: dict[str, Any] = {}


class RunRecord(BaseModel):
    run_id: str | None = None
    endpoint: str
    workflow_name: str | None = None
    workflow_version: str | None = None
    workflow_id: str | None = None
    status: str | None = None
    submitted_at: str
    started_at: str | None = None
    executions: list[WorkflowExecution] | None = None
    error: str | None = None



from examples.structured_demo.workflow.builder import build_demo_workflow
from examples.structured_demo.workflow.input import DemoWorkflowInput
from fluxly import Fluxly


def main() -> None:
    wf = build_demo_workflow()
    cli = Fluxly()
    cli.add_endpoint("structured-demo", wf, DemoWorkflowInput)
    cli.run()


if __name__ == "__main__":
    main()






import time
import unittest

from fluxly.exceptions import DataErrorException
from fluxly.node import Node
from fluxly.status import StatusCodes
from fluxly.workflow import Workflow, WorkflowInput


class SlowOk(Node):
    sleep_seconds: float = 0.3
    def _logic(self) -> None:
        time.sleep(self.sleep_seconds)


class FailFast(Node):
    def _logic(self) -> None:
        raise DataErrorException()


class ExecutionGroupsTest(unittest.TestCase):
    def _wf(self) -> Workflow:
        return Workflow(name="eg-wf", description="exec groups", version="v", inputs=WorkflowInput(verbose=False))

    def test_partial_success_across_groups_completes_workflow(self) -> None:
        wf = self._wf()

        slow_ok = SlowOk(name="slow-ok", sleep_seconds=0.4)
        fail_fast = FailFast(name="fail-fast")
        ok2 = SlowOk(name="ok2", sleep_seconds=0.05)

        wf.add_node(slow_ok)
        wf.add_node(fail_fast)
        wf.add_node(ok2)

        # Define execution groups
        wf.add_execution_group([slow_ok, fail_fast])
        wf.add_execution_group([ok2])

        # Ensure group 2 runs only after slow_ok completes successfully
        wf._graph.add_edge_if_source_completed(slow_ok, ok2)

        wf.execute()

        # Workflow should complete even if one node in Group 1 failed,
        # because Group 2 has a successful node and not all groups are "dead".
        self.assertEqual(wf.last_execution.status, StatusCodes.COMPLETED)

        # Assert node outcomes
        self.assertEqual(slow_ok.last_execution.status, StatusCodes.COMPLETED)
        self.assertEqual(ok2.last_execution.status, StatusCodes.COMPLETED)
        self.assertNotEqual(fail_fast.last_execution.status, StatusCodes.COMPLETED)

    def test_all_groups_have_failure_causes_workflow_failure(self) -> None:
        wf = self._wf()

        slow_ok_1 = SlowOk(name="slow-ok-1", sleep_seconds=0.3)
        fail_fast_1 = FailFast(name="fail-fast-1")
        fail_fast_2 = FailFast(name="fail-fast-2")

        wf.add_node(slow_ok_1)
        wf.add_node(fail_fast_1)
        wf.add_node(fail_fast_2)

        # Two groups, each will have a failing node
        wf.add_execution_group([slow_ok_1, fail_fast_1])
        wf.add_execution_group([fail_fast_2])

        # Delay the second group's start until slow_ok_1 completes,
        # to make failure ordering deterministic
        wf._graph.add_edge_if_source_completed(slow_ok_1, fail_fast_2)

        with self.assertRaises(DataErrorException):
            wf.execute()

        # Workflow status should reflect failure
        self.assertEqual(wf.last_execution.status, StatusCodes.DATA_ERROR)


if __name__ == "__main__":
    unittest.main()





from typing import Annotated

from pydantic import BaseModel, Field


class LoggerConfig(BaseModel):
    level: Annotated[str, Field("INFO", description="log level: DEBUG | INFO | WARNING | ERROR | FATAL")]
    file: Annotated[str, Field(None, description="log output file")]
    max_print_length: int = Field(10*1000, description="Maximum length of log messages before truncation")



from fluxly.core.node.error import NodeError
from fluxly.core.node.execution import NodeExecution
from fluxly.core.node.metadata import NodeMetadata
from fluxly.core.node.node import Node
from fluxly.core.node.output import NodeOutput

__all__ = [
    "Node",
    "NodeMetadata",
    "NodeOutput",
    "NodeExecution",
    "NodeError",
]



import copy
from typing import Any

import click
from pydantic import ValidationError

from fluxly.core.utils.consts import ENV_PREFIX
from fluxly.core.workflow.input import WorkflowInput
from fluxly.core.workflow.models import EndpointType
from fluxly.core.workflow.workflow import Workflow


def _json_type_to_click_type(json_type: str) -> Any:
    mapping: dict[str, Any] = {"integer": int, "number": float, "string": str}
    return mapping.get(json_type, str)


def _resolve_field_type(meta: dict[str, Any] | None) -> str:
    if not meta:
        return "string"
    t = meta.get("type")
    if t:
        return t
    any_of = meta.get("anyOf") or meta.get("oneOf") or []
    for variant in any_of if isinstance(any_of, list) else []:
        if isinstance(variant, dict) and variant.get("type") and variant.get("type") != "null":
            return variant.get("type")  # type: ignore[return-value]
    return "string"


def _get_properties_and_required(workflow_input_cls: type[WorkflowInput]) -> tuple[dict[str, Any], set[str]]:
    schema: dict[str, Any] = workflow_input_cls.model_json_schema()
    return schema.get("properties", {}), set(schema.get("required", []))


def _is_excluded_from_cli(meta: dict[str, Any]) -> bool:
    extra = meta.get("json_schema_extra", {}) if isinstance(meta, dict) else {}
    return bool(extra.get("exclude_from_cli"))


def _cli_name(field_name: str, meta: dict[str, Any]) -> str:
    extra = meta.get("json_schema_extra", {}) if isinstance(meta, dict) else {}
    return field_name.replace("_", "-") if extra.get("convert_underscores", True) else field_name


def _build_option(field_name: str, meta: dict[str, Any], is_required: bool) -> click.Option:
    cli = _cli_name(field_name, meta)
    help_text = meta.get("description", "")
    default = meta.get("default")
    enum_values = meta.get("enum")
    field_type = _resolve_field_type(meta)
    envvar_name = f"{ENV_PREFIX}{cli.upper().replace('-', '_')}"

    if field_type == "boolean":
        opt = click.Option(
            [f"--{cli}/--no-{cli}"],
            is_flag=True,
            default=bool(default) if default is not None else False,
            help=help_text,
            envvar=envvar_name,
            show_envvar=True,
        )
    elif field_type == "array":
        elem_type: Any = str
        items_meta = meta.get("items") if isinstance(meta.get("items"), dict) else None
        if items_meta:
            elem_type = _json_type_to_click_type(_resolve_field_type(items_meta))  # type: ignore[arg-type]
        opt = click.Option(
            [f"--{cli}"],
            multiple=True,
            type=elem_type,
            required=is_required,
            help=help_text,
            default=tuple(default) if isinstance(default, list) else (),
            show_default=bool(default),
            envvar=envvar_name,
            show_envvar=True,
            split_envvar=True,
        )
    else:
        click_type: Any = click.Choice([str(v) for v in enum_values]) if enum_values else _json_type_to_click_type(field_type)
        opt = click.Option(
            [f"--{cli}"],
            type=click_type,
            required=is_required,
            default=default,
            help=help_text,
            show_default=default is not None,
            envvar=envvar_name,
            show_envvar=True,
        )

    opt.name = field_name
    return opt


def _build_click_params(workflow_input_cls: type[WorkflowInput]) -> list[click.Parameter]:
    properties, required = _get_properties_and_required(workflow_input_cls)
    params: list[click.Parameter] = []
    for name, meta_any in properties.items():
        meta: dict[str, Any] = meta_any if isinstance(meta_any, dict) else {}
        if _is_excluded_from_cli(meta):
            continue
        is_required = name in required and meta.get("default") is None
        params.append(_build_option(name, meta, is_required))
    return params


def _normalize_array_options(kwargs: dict[str, Any], workflow_input_cls: type[WorkflowInput]) -> dict[str, Any]:
    properties, _ = _get_properties_and_required(workflow_input_cls)
    normalized: dict[str, Any] = {}
    for key, value in kwargs.items():
        meta: dict[str, Any] = properties.get(key, {}) if isinstance(properties.get(key, {}), dict) else {}
        if _resolve_field_type(meta) == "array" and isinstance(value, tuple):
            normalized[key] = list(value)
        else:
            normalized[key] = value
    return normalized


def _create_inputs(workflow_input_cls: type[WorkflowInput], values: dict[str, Any]) -> WorkflowInput:
    try:
        return workflow_input_cls(**values)
    except ValidationError as e:
        raise click.ClickException(f"Input validation failed:\n{e}")


def _clone_workflow(template: Workflow) -> Workflow:
    return copy.deepcopy(template)


def build_click_command_for_workflow(command_name: str, workflow_template: Workflow, workflow_input_cls: type[WorkflowInput]) -> click.Command:
    params = _build_click_params(workflow_input_cls)

    def _callback(**kwargs: Any) -> None:
        normalized = _normalize_array_options(kwargs, workflow_input_cls)
        inputs = _create_inputs(workflow_input_cls, normalized)
        wf = _clone_workflow(workflow_template)
        wf.inputs = inputs
        wf.assign_trigger(endpoint_type=EndpointType.CLI, endpoint_name=command_name)
        wf.init_by_cli()

    return click.Command(name=command_name, params=params, callback=_callback, help=f"Execute the {command_name} workflow.")



def build_click_group_with_commands(commands: dict[str, tuple[Workflow, type[WorkflowInput]]]) -> click.Group:
    group = click.Group(help="Fluxly Framework Runner")
    for command_name, (workflow_template, wf_input_cls) in commands.items():
        group.add_command(build_click_command_for_workflow(command_name, workflow_template, wf_input_cls), name=command_name)
    return group




import time
import unittest

from fluxly.exceptions import (
    DataErrorException,
    TimeoutException,
)
from fluxly.node import Node
from fluxly.status import StatusCodes
from fluxly.workflow import Workflow, WorkflowInput


class OkNode(Node):
    def _logic(self) -> None:
        pass


class FailNode(Node):
    def _logic(self) -> None:
        raise DataErrorException("failure")


class SlowNode(Node):
    def _logic(self) -> None:
        time.sleep((self.timeout_seconds or 1) + 1)


class WorkflowBehaviorTest(unittest.TestCase):
    def _wf(self, node: Node, timeout_seconds: int | None = None) -> Workflow:
        inputs = WorkflowInput(verbose=False, timeout_seconds=timeout_seconds)
        wf = Workflow(
            name="test-workflow",
            description="test workflow",
            version="1.0",
            inputs=inputs
        )
        wf.add_node(node)
        return wf

    def test_success(self) -> None:
        wf = self._wf(OkNode(name="ok-node"))
        wf.execute()
        self.assertEqual(wf.last_execution.status, StatusCodes.COMPLETED)

    def test_retry_then_success(self) -> None:
        class Flaky(Node):
            fail_times: int = 1
            def _logic(self) -> None:
                if self.attempt <= self.fail_times:
                    raise DataErrorException("transient")

        node = Flaky(name="flaky", max_retries=2, retry_delay_seconds=0)
        wf = self._wf(node)
        wf.execute()
        self.assertEqual(node.attempt, 2)
        self.assertEqual(wf.last_execution.status, StatusCodes.COMPLETED)

    def test_failure_sets_status_and_raises(self) -> None:
        wf = self._wf(FailNode(name="fail"))
        with self.assertRaises(DataErrorException):
            wf.execute()
        self.assertEqual(wf.last_execution.status, StatusCodes.DATA_ERROR)

    def test_timeout_sets_status_and_raises(self) -> None:
        node = SlowNode(name="slow", timeout_seconds=1)
        wf = self._wf(node, timeout_seconds=1)
        with self.assertRaises(TimeoutException):
            wf.execute()
        self.assertEqual(wf.last_execution.status, StatusCodes.TIMED_OUT)


if __name__ == "__main__":
    unittest.main()





from datetime import datetime, timedelta
from typing import Annotated

from pydantic import PlainSerializer

from fluxly.core.utils.formatting import (
    format_datetime_to_humanreadable,
    format_timedelta_to_humanreadable,
)

DatetimeReadable = Annotated[
    datetime | None,
    PlainSerializer(lambda v: format_datetime_to_humanreadable(v)),
]

TimedeltaReadable = Annotated[
    timedelta | None,
    PlainSerializer(lambda v: format_timedelta_to_humanreadable(v)),
]



from __future__ import annotations

from typing import Any

from fastapi import HTTPException, status
from fastapi.responses import JSONResponse
from pydantic import BaseModel, ValidationError

from fluxly.core.api.service import RunnerService
from fluxly.core.workflow.input import WorkflowInput
from fluxly.core.workflow.workflow import Workflow


class EndpointRunner(BaseModel):
    model_config = {"arbitrary_types_allowed": True}
    name: str
    workflow: Workflow
    input_cls: type[WorkflowInput]
    service: RunnerService

    async def submit(self, payload: dict[str, Any]) -> JSONResponse:
        values = payload
        try:
            self.input_cls(**values)
        except (ValidationError) as e:
            raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_CONTENT, detail=str(e))

        record = self.service.submit(self.name, self.workflow, self.input_cls, values)
        return JSONResponse(status_code=status.HTTP_202_ACCEPTED, content=record.model_dump())


def get_run_handler(service: RunnerService):
    async def _get_run(run_id: str) -> Any:
        record = service.get(run_id)
        if not record:
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="Run not found")
        return record.model_dump()

    return _get_run


def get_run_by_endpoint_handler(service: RunnerService):
    async def _get_run_by_endpoint(endpoint: str, run_id: str) -> Any:
        record = service.get(run_id)
        if not record or record.endpoint != endpoint:
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="Run not found")
        return record.model_dump()

    return _get_run_by_endpoint


async def health_handler() -> dict[str, str]:
    return {"status": "ok"}





import unittest

from fluxly.node import Node
from fluxly.workflow import WorkflowGraph


class NodeExecutionError(Exception):
    pass

class DummyNode(Node):
    fail: bool = False

    def _logic(self) -> None:
        if self.fail:
            raise NodeExecutionError(f"{self.name} failed")
        print(f"{self.name} executed successfully")


class WorkflowGraphTest(unittest.TestCase):
    def setUp(self) -> None:
        self.node_a = DummyNode(name="test-A")
        self.node_b = DummyNode(name="test-B")
        self.node_c = DummyNode(name="test-C")
        self.graph = WorkflowGraph()
        self.graph.add_node(self.node_a)
        self.graph.add_node(self.node_b)
        self.graph.add_node(self.node_c)

    def test_simple_dependency(self) -> None:
        self.graph.add_edge(self.node_a, self.node_b)
        completed = set()
        self.assertTrue(self.graph.can_node_run(self.node_a, completed))
        self.assertFalse(self.graph.can_node_run(self.node_b, completed))
        self.node_a.execute()
        completed.add(self.node_a)
        self.assertTrue(self.graph.can_node_run(self.node_b, completed))

    def test_acyclic_enforcement(self) -> None:
        self.graph.add_edge(self.node_a, self.node_b)
        self.graph.add_edge(self.node_b, self.node_c)
        with self.assertRaises(ValueError):
            self.graph.add_edge(self.node_c, self.node_a)

    def test_duplicate_node_rejected(self) -> None:
        with self.assertRaises(ValueError):
            self.graph.add_node(DummyNode(name="test-A"))

    def test_duplicate_edge_rejected(self) -> None:
        self.graph.add_edge(self.node_a, self.node_b)
        with self.assertRaises(ValueError):
            self.graph.add_edge(self.node_a, self.node_b)

    def test_conditional_edge(self) -> None:
        def condition():
            return True
        self.graph.add_conditional_edge(self.node_a, self.node_b, condition)
        completed = set()
        self.assertTrue(self.graph.can_node_run(self.node_a, completed))
        self.assertFalse(self.graph.can_node_run(self.node_b, completed))
        self.node_a.execute()
        completed.add(self.node_a)
        self.assertTrue(self.graph.can_node_run(self.node_b, completed))

    def test_failed_node(self) -> None:
        fail_node = DummyNode(name="Fail", fail=True)
        self.graph.add_node(fail_node)
        self.graph.add_conditional_edge(fail_node, self.node_c, condition=lambda: True)
        completed = set()
        with self.assertRaises(NodeExecutionError):
            fail_node.execute()
        completed.add(fail_node) # Regardless of the failure, the node should be considered completed
        self.assertTrue(self.graph.can_node_run(self.node_c, completed)) # The node should be able to run because the condition is met

    def test_edge_if_source_completed(self) -> None:
        self.graph.add_edge_if_source_completed(self.node_a, self.node_b)
        completed = set()
        self.assertTrue(self.graph.can_node_run(self.node_a, completed))
        self.assertFalse(self.graph.can_node_run(self.node_b, completed))

        self.node_a.execute()
        completed.add(self.node_a)
        self.assertTrue(self.graph.can_node_run(self.node_b, completed))


if __name__ == "__main__":
    unittest.main()



from fluxly.services.logger.config import LoggerConfig
from fluxly.services.logger.logger import LoggerService
from fluxly.services.services import Services

__all__ = [
    "LoggerConfig",
    "LoggerService",
    "Services",
]



import os
import sys
import unittest

from fluxly import Fluxly
from fluxly.exceptions import DataErrorException
from fluxly.node import Node
from fluxly.status import StatusCodes
from fluxly.workflow import Workflow, WorkflowInput


class OkNode(Node):
    def _logic(self) -> None:
        pass


class FailNode(Node):
    def _logic(self) -> None:
        raise DataErrorException()


def _build_cli(cmd_name: str, node: Node) -> Fluxly:
    wf = Workflow(name="test-cli-wf", description="cli test workflow", inputs=WorkflowInput(verbose=False))
    wf.add_node(node)
    cli = Fluxly()
    cli.add_command(cmd_name, wf, WorkflowInput)
    return cli


class FluxlyCliTest(unittest.TestCase):
    def setUp(self) -> None:
        self._orig_argv = list(sys.argv)

    def tearDown(self) -> None:
        sys.argv = self._orig_argv

    def test_cli_success_exit_code_zero(self) -> None:
        cli = _build_cli("ok", OkNode(name="ok-node"))
        sys.argv = ["prog", "ok"]
        with self.assertRaises(SystemExit) as ctx:
            cli.run_cli()
        self.assertEqual(ctx.exception.code, 0)


    def test_cli_failure_exit_code_custom(self) -> None:
        cli = _build_cli("fail", FailNode(name="fail-node"))
        sys.argv = ["prog", "fail"]
        with self.assertRaises(SystemExit) as ctx:
            cli.run_cli()
        self.assertEqual(ctx.exception.code, StatusCodes.DATA_ERROR.value)



    def test_cli_env_only_params_are_applied(self) -> None:
        class AssertInputsNode(Node):
            def _logic(self) -> None:
                assert self._workflow_input is not None
                assert self._workflow_input.max_retries == 2
                assert self._workflow_input.retry_delay_seconds == 1
                assert self._workflow_input.timeout_seconds == 5
                assert self._workflow_input.md_file_path == "env_doc.md"
                assert self._workflow_input.diagram_file_path == "env_diagram.png"

        cli = _build_cli("ok", AssertInputsNode(name="assert-inputs"))

        prev_env = {
            "FLUXLY_MAX_RETRIES": os.environ.get("FLUXLY_MAX_RETRIES"),
            "FLUXLY_RETRY_DELAY_SECONDS": os.environ.get("FLUXLY_RETRY_DELAY_SECONDS"),
            "FLUXLY_TIMEOUT_SECONDS": os.environ.get("FLUXLY_TIMEOUT_SECONDS"),
            "FLUXLY_MD_FILE_PATH": os.environ.get("FLUXLY_MD_FILE_PATH"),
            "FLUXLY_DIAGRAM_FILE_PATH": os.environ.get("FLUXLY_DIAGRAM_FILE_PATH"),
        }

        os.environ["FLUXLY_MAX_RETRIES"] = "2"
        os.environ["FLUXLY_RETRY_DELAY_SECONDS"] = "1"
        os.environ["FLUXLY_TIMEOUT_SECONDS"] = "5"
        os.environ["FLUXLY_MD_FILE_PATH"] = "env_doc.md"
        os.environ["FLUXLY_DIAGRAM_FILE_PATH"] = "env_diagram.png"

        try:
            sys.argv = ["prog", "ok"]
            with self.assertRaises(SystemExit) as ctx:
                cli.run_cli()
            self.assertEqual(ctx.exception.code, 0)
        finally:
            for k, v in prev_env.items():
                if v is None:
                    os.environ.pop(k, None)
                else:
                    os.environ[k] = v




from fluxly.core.exceptions import WorkflowException
from fluxly.core.status import StatusCodes


class NodesNotFoundException(WorkflowException):
    exit_code = StatusCodes.PREREQUISITE_FAIL


class UnsupportedGraphScenario(WorkflowException):
    exit_code = StatusCodes.PREREQUISITE_FAIL



from typing import Annotated

from pydantic import BaseModel, Field

from fluxly.core.node.execution import NodeExecution


class WorkflowOutput(BaseModel):
    node_to_executions: Annotated[dict[str, list[NodeExecution]], Field(description="Mapping of node name to all its executions")] = {}

    def __str__(self) -> str:
        return self.model_dump_json(indent=2)

    def __repr__(self) -> str:
        return self.__str__()



from typing import Annotated

from pydantic import Field

from fluxly.core.workflow.input import WorkflowInput as CoreWorkflowInput


class DemoWorkflowInput(CoreWorkflowInput):
    message: Annotated[str | None, Field(description="Optional message passed to Alpha")] = None
    suffix: Annotated[str | None, Field(description="Optional suffix passed to Beta")] = None






from typing import ClassVar

from pydantic import BaseModel

from fluxly.services.logger.config import LoggerConfig
from fluxly.services.logger.logger import LoggerService


class Services(BaseModel):
    _logger_instance: ClassVar[LoggerService] = None

    @classmethod
    def get_logger(cls) -> LoggerService:
        if cls._logger_instance is None:
            cls._logger_instance = LoggerService(
                config=LoggerConfig(),
            )
        return cls._logger_instance



import time
import unittest

from fluxly.exceptions import DataErrorException, TimeoutException
from fluxly.node import Node
from fluxly.status import StatusCodes


class FlakyNode(Node):
    fail_times: int = 1

    def _logic(self) -> None:
        if self.attempt <= self.fail_times:
            raise DataErrorException("transient failure")


class AlwaysFailNode(Node):
    def _logic(self) -> None:
        raise DataErrorException("permanent failure")


class TimeoutNode(Node):
    def _logic(self) -> None:
        time.sleep((self.timeout_seconds or 1) + 1)


class NodeExecutionTest(unittest.TestCase):
    def test_retry_success_then_complete(self) -> None:
        node = FlakyNode(name="flaky", fail_times=1, max_retries=2, retry_delay_seconds=1)
        node.execute()

        self.assertEqual(node.attempt, 2)
        self.assertEqual(len(node.executions), 2)

        first, second = node.executions[0], node.executions[1]
        self.assertEqual(first.status, StatusCodes.DATA_ERROR)
        self.assertIsNotNone(first.error)
        self.assertEqual(second.status, StatusCodes.COMPLETED)

    def test_retry_exhausted_raises(self) -> None:
        node = AlwaysFailNode(name="always-fail", max_retries=3, retry_delay_seconds=1)
        with self.assertRaises(DataErrorException):
            node.execute()

        self.assertEqual(node.attempt, 3)
        self.assertEqual(len(node.executions), 3)
        for ex in node.executions:
            self.assertEqual(ex.status, StatusCodes.DATA_ERROR)
            self.assertIsNotNone(ex.error)

    def test_timeout_sets_status_and_raises(self) -> None:
        node = TimeoutNode(name="timeout", timeout_seconds=1, max_retries=1)
        with self.assertRaises(TimeoutException):
            node.execute()

        self.assertEqual(node.attempt, 1)
        self.assertEqual(len(node.executions), 1)
        latest = node.executions[-1]
        self.assertEqual(latest.status, StatusCodes.TIMED_OUT)
        self.assertIsNotNone(latest.error)
        self.assertEqual(latest.error.exception_class_name, "TimeoutException")


if __name__ == "__main__":
    unittest.main()





import time

from fluxly.node import Node
from fluxly.workflow import Workflow, WorkflowInput


class NodeA(Node):
    def _logic(self) -> None:
        time.sleep(0.2)
        _demo_state["do_c"] = True


class NodeB(Node):
    def _logic(self) -> None:
        time.sleep(0.1)


class NodeC(Node):
    def _logic(self) -> None:
        time.sleep(0.1)

wf = Workflow(name="demo", description="Demo concurrent workflow", version="0.1")
a = NodeA(name="test A node", description="Root node", timeout_seconds=5)
b = NodeB(name="test B node", description="Runs after A", timeout_seconds=5, max_retries=2)
c = NodeC(name="test C node", description="Conditionally runs after A", timeout_seconds=5)
d = NodeC(name="test d node", description="runs after c", timeout_seconds=5)

_demo_state = {"do_c": False}


def build_demo_workflow() -> Workflow:
    wf.add_node(a)
    wf.add_node(b)
    wf.add_node(c)
    wf.add_node(d)

    wf.add_edge(a, b)
    wf.add_conditional_edge(a, c, condition=lambda: _demo_state["do_c"])
    wf.add_edge(c, d)
    return wf


if __name__ == "__main__":
    workflow = build_demo_workflow()
    workflow.inputs = WorkflowInput(verbose=True, timeout_seconds=30, max_retries=1, retry_delay_seconds=0, auto_generate_md=True)

    workflow.add_execution_group([a, b])
    workflow.execute()

    print(f"Workflow status: {workflow.last_execution.status} ")
    print(f"Workflow output: ({workflow.last_execution.output})")




from pydantic import BaseModel


class NodeOutput(BaseModel):
    def __str__(self) -> str:
        return self.model_dump_json(indent=2)

    def __repr__(self) -> str:
        return self.__str__()



from typing import Annotated

from pydantic import Field

from fluxly.node import NodeOutput


class BetaOutput(NodeOutput):
    combined: Annotated[str, Field(description="Output combined message")] = ""



from collections.abc import Callable
from graphlib import CycleError, TopologicalSorter

from pydantic import BaseModel, Field, PrivateAttr

from fluxly.core.node.node import Node
from fluxly.core.status import StatusCodes


class Edge(BaseModel):
    source: str = Field(...)
    destination: str = Field(...)
    condition: Callable[[], bool] | None = None
    condition_passed: bool | None = None

    def evaluate_condition(self) -> bool:
        self.condition_passed = True if self.condition is None else bool(self.condition())
        return self.condition_passed


class WorkflowGraph(BaseModel):
    edges: list[Edge] = Field(default_factory=list)
    _nodes: dict[str, Node] = PrivateAttr(default_factory=dict)

    @property
    def nodes(self) -> dict[str, Node]:
        # Return the live mapping of node instances (by reference)
        return self._nodes

    def get_parents(self, node: Node) -> list[Node]:
        return [self._nodes[edge.source] for edge in self.edges if edge.destination == node.name]

    def get_children(self, node: Node) -> list[Node]:
        return [self._nodes[edge.destination] for edge in self.edges if edge.source == node.name]

    def add_node(self, node: Node) -> None:
        if node.name in self._nodes:
            raise ValueError(f"Node '{node.name}' already exists.")
        self._nodes[node.name] = node

    def add_edge(self, source_node: Node, dest_node: Node) -> Edge:
        self._validate_nodes(source_node, dest_node)
        self._validate_no_duplicate_edge(source_node, dest_node)
        edge = Edge(source=source_node.name, destination=dest_node.name)
        self._validate_acyclic(edge)
        self.edges.append(edge)
        return edge

    def add_conditional_edge(self, source_node: Node, dest_node: Node, condition: Callable[[], bool]) -> Edge:
        self._validate_nodes(source_node, dest_node)
        self._validate_no_duplicate_edge(source_node, dest_node)
        edge = Edge(source=source_node.name, destination=dest_node.name, condition=condition)
        self._validate_acyclic(edge)
        self.edges.append(edge)
        return edge

    def add_edge_if_source_completed(self, source_node: Node, dest_node: Node) -> Edge:
        self._validate_nodes(source_node, dest_node)
        self._validate_no_duplicate_edge(source_node, dest_node)

        def condition() -> bool:
            return source_node.attempt > 0 and source_node.last_execution.status == StatusCodes.COMPLETED

        edge = Edge(source=source_node.name, destination=dest_node.name, condition=condition)
        self._validate_acyclic(edge)
        self.edges.append(edge)
        return edge

    def get_edge(self, source: Node, destination: Node) -> Edge | None:
        return next(
            (edge for edge in self.edges if edge.source == source.name and edge.destination == destination.name),
            None,
        )

    def _validate_nodes(self, source_node: Node, dest_node: Node) -> None:
        if source_node.name not in self.nodes or dest_node.name not in self.nodes:
            raise ValueError("Both source and destination nodes must exist in the graph.")
        if source_node.name == dest_node.name:
            raise ValueError("Cannot create self-loop edges.")

    def _validate_acyclic(self, new_edge: Edge) -> None:
        sorter = TopologicalSorter()

        for e in self.edges:
            sorter.add(e.destination, e.source)

        sorter.add(new_edge.destination, new_edge.source)

        try:
            list(sorter.static_order())
        except CycleError:
            raise ValueError(f"edge {new_edge.source} → {new_edge.destination} creates a cycle.")

    def _edge_exists(self, source: str, destination: str) -> bool:
        return any(e.source == source and e.destination == destination for e in self.edges)

    def _validate_no_duplicate_edge(self, source_node: Node, dest_node: Node) -> None:
        if self._edge_exists(source_node.name, dest_node.name):
            raise ValueError(f"Edge '{source_node.name}' → '{dest_node.name}' already exists.")

    def can_node_run(self, node: Node, completed: set[Node]) -> bool:
        if node.name in completed:
            return False

        parents = self.get_parents(node)

        if not parents:
            return True

        completed_names = {n.name for n in completed}

        for parent in parents:
            if parent.name not in completed_names:
                return False

            edge = self.get_edge(parent, node)
            if edge is not None and edge.condition is not None:
                if not edge.condition():
                    return False

        return True



import unittest

from fluxly.node import Node, NodeExecution, NodeOutput
from fluxly.status import StatusCodes
from fluxly.workflow import Workflow, WorkflowInput


class ProducerOutput(NodeOutput):
    value: str = ""


class ProducerExecution(NodeExecution):
    output: ProducerOutput = ProducerOutput()


class Producer(Node):
    produced_value: str = "alpha"

    def _create_execution(self) -> ProducerExecution:
        return ProducerExecution()

    def _logic(self) -> None:
        self.current_execution.output.value = self.produced_value


class ConsumerOutput(NodeOutput):
    seen: str = ""


class ConsumerExecution(NodeExecution):
    output: ConsumerOutput = ConsumerOutput()


class Consumer(Node):
    producer_node: Producer | None = None

    def _create_execution(self) -> ConsumerExecution:
        return ConsumerExecution()

    def _logic(self) -> None:
        assert self.producer_node is not None, "Producer node must be provided before running Consumer"
        self.current_execution.output.seen = self.producer_node.last_execution.output.value


class Node2NodeCommunicationTest(unittest.TestCase):
    def _wf(self, nodes: list[Node]) -> Workflow:
        wf = Workflow(
            name="n2n",
            description="node to node",
            inputs=WorkflowInput(verbose=False),
        )
        for n in nodes:
            wf.add_node(n)
        return wf

    def test_consumer_reads_actual_producer_execution_value(self) -> None:
        producer = Producer(name="producer", produced_value="from-execution")
        consumer = Consumer(name="consumer", producer_node=producer)

        wf = self._wf([producer, consumer])
        wf.add_edge_if_source_completed(producer, consumer)
        wf.execute()

        self.assertEqual(producer.last_execution.status, StatusCodes.COMPLETED)
        self.assertEqual(consumer.last_execution.status, StatusCodes.COMPLETED)
        self.assertEqual(consumer.last_execution.output.seen, "from-execution")
        self.assertIs(consumer.producer_node, producer)
        consumer.producer_node.last_execution.output.value = "mutated"
        self.assertEqual(producer.last_execution.output.value, "mutated")


if __name__ == "__main__":
    unittest.main()






from examples.structured_demo.nodes.alpha.execution import AlphaNodeExecution
from examples.structured_demo.workflow.input import DemoWorkflowInput
from fluxly.core.node.node import Node


class Alpha(Node):
    _workflow_input: DemoWorkflowInput

    @property
    def current_execution(self) -> AlphaNodeExecution:
        return super().current_execution

    @property
    def last_execution(self) -> AlphaNodeExecution:
        return super().last_execution

    @property
    def workflow_input(self) -> DemoWorkflowInput:
        return self._workflow_input

    def _create_execution(self) -> AlphaNodeExecution:
        return AlphaNodeExecution()

    def _logic(self) -> None:
        msg = self.workflow_input.message
        if not msg:
            msg = "hello from alpha"

        self.current_execution.output.echoed = msg
        self._logger.info(f"Alpha produced: {self.current_execution.output.echoed}")





import sys
import threading
import time
from collections.abc import Callable
from datetime import datetime
from typing import Annotated
from uuid import uuid4

from pydantic import BaseModel, Field, PrivateAttr

from fluxly.core.docs_generator.generator import generate_workflow_documentation
from fluxly.core.exceptions import TimeoutException, WorkflowException
from fluxly.core.node.node import Node
from fluxly.core.status import StatusCodes
from fluxly.core.workflow.exceptions import (
    NodesNotFoundException,
    UnsupportedGraphScenario,
)
from fluxly.core.workflow.execution import WorkflowExecution
from fluxly.core.workflow.graph import WorkflowGraph
from fluxly.core.workflow.input import WorkflowInput
from fluxly.core.workflow.metadata import WorkflowMetadata
from fluxly.core.workflow.models import EndpointType
from fluxly.core.workflow.utils import build_cli_command_from_workflow_input
from fluxly.services import LoggerConfig, LoggerService


class Workflow(BaseModel):
    name: Annotated[str, Field(max_length=64, min_length=2, description="The name of the workflow.")]
    description: Annotated[str | None, Field(max_length=128, min_length=3, description="The description of the workflow.")] = None
    version: Annotated[str | None, Field(description="The workflow version.")] = None
    inputs: Annotated[WorkflowInput | None, Field(description="The workflow inputs.")] = None

    _id: str = PrivateAttr(default_factory=lambda: str(uuid4()))
    _run_id: str | None = PrivateAttr(default=None)
    _executions: list[WorkflowExecution] = PrivateAttr(default_factory=list)
    _graph: WorkflowGraph = PrivateAttr(default_factory=WorkflowGraph)
    _execution_groups: list[set[str]] = PrivateAttr(default_factory=list)
    _logger: LoggerService = PrivateAttr(default_factory=lambda: LoggerService(config=LoggerConfig()))
    _endpoint_type: EndpointType | None = PrivateAttr(default=None)
    _endpoint_name: str | None = PrivateAttr(default=None)

    @property
    def id(self) -> str:
        return self._id

    @property
    def run_id(self) -> str:
        return self._run_id

    def assign_run_id(self, run_id: str) -> None:
        self._run_id = run_id

    def assign_trigger(self, endpoint_type: EndpointType, endpoint_name: str) -> None:
        self._endpoint_type = endpoint_type
        self._endpoint_name = endpoint_name

    @property
    def endpoint_type(self) -> EndpointType | None:
        return self._endpoint_type

    @property
    def endpoint_name(self) -> str | None:
        return self._endpoint_name

    @property
    def current_execution(self) -> WorkflowExecution:
        return self._executions[-1]

    @property
    def last_execution(self) -> WorkflowExecution:
        return self._executions[-1]

    @property
    def attempt(self) -> int:
        return len(self._executions)

    @property
    def metadata(self) -> WorkflowMetadata:
        return self.current_execution.metadata

    @property
    def executions(self) -> list[WorkflowExecution]:
        return self._executions

    def init_by_cli(self) -> None:
        try:
            self.execute()
        finally:
            if self.last_execution:
                sys.exit(self.last_execution.status.value)
            else:
                sys.exit(1)

    def add_node(self, node: Node) -> None:
        self._graph.add_node(node)

    def add_edge(self, source_node: Node, dest_node: Node) -> None:
        self._graph.add_edge(source_node, dest_node)

    def add_conditional_edge(self, source_node: Node, dest_node: Node, condition: Callable[[], bool]) -> None:
        self._graph.add_conditional_edge(source_node, dest_node, condition)

    def add_edge_if_source_completed(self, source_node: Node, dest_node: Node) -> None:
        self._graph.add_edge_if_source_completed(source_node, dest_node)

    def add_execution_group(self, nodes: list[Node]) -> None:
        if not nodes:
            raise ValueError("Execution group must include at least one node.")

        node_names = {node.name for node in nodes}
        missing_nodes = node_names - self._graph.nodes.keys()
        if missing_nodes:
            raise ValueError(f"Execution group contains unknown nodes: {sorted(missing_nodes)}")

        self._execution_groups.append(node_names)

    def get_nodes(self) -> list[Node]:
        return list(self._graph.nodes.values())

    def get_execution_groups(self) -> list[set[str]]:
        return [group.copy() for group in self._execution_groups]

    def _create_execution(self) -> WorkflowExecution:
        return WorkflowExecution(id=str(self.attempt + 1))

    def _all_execution_groups_dead(self) -> bool:
        nodes = self._graph.nodes
        groups = self._execution_groups or [set(nodes.keys())]

        for group in groups:
            group_failed = any(
                (nodes[n].attempt > 0 and nodes[n].last_execution.status != StatusCodes.COMPLETED)
                for n in group if n in nodes
            )
            if not group_failed:
                return False

        return True

    def execute(self) -> None:
        if not self._run_id:
            self._run_id = str(uuid4())
        try:
            if not self._graph.nodes:
                raise NodesNotFoundException

            self._log_workflow_start()
            while self.attempt <= self.inputs.max_retries:
                try:
                    self._start_workflow_execution()
                    self.on_start()

                    self._run_with_timeout()
                    self.on_success()
                    break
                except Exception as e:
                    self.on_failure(e)
                    if not self._handle_retry(e):
                        raise e
                finally:
                    self._finalize_workflow_execution()
                    self.on_finish()

        finally:
            self._finalize_workflow()

    def _start_workflow_execution(self) -> None:
        execution = self._create_execution()
        execution._attempt = self.attempt + 1
        execution.metadata.start_time = datetime.now()
        execution.status = StatusCodes.IN_PROGRESS
        self._executions.append(execution)

    def _run_with_timeout(self) -> None:
        result: list[Exception | bool] = []

        def runner() -> None:
            try:
                self._iterate_nodes()
                result.append(True)
            except Exception as e:
                result.append(e)

        thread = threading.Thread(target=runner, daemon=True)
        thread.start()
        thread.join(timeout=self.inputs.timeout_seconds)

        if thread.is_alive():
            self._handle_exception(TimeoutException())

        if result and isinstance(result[0], Exception):
            self._handle_exception(result[0])

    def _handle_exception(self, exception: Exception | WorkflowException | None) -> None:
        if exception is None:
            return
        current = self.current_execution
        status = exception.exit_code if isinstance(exception, WorkflowException) else StatusCodes.FAILED
        current.status = status
        raise exception

    def _handle_retry(self, error: Exception) -> bool:
        if self.attempt >= self.inputs.max_retries:
            self._logger.error(f"{self.name} failed: {error}. Retries exhausted.")
            return False

        self._logger.warning(f"{self.name} failed: {error}. Retrying in {self.inputs.retry_delay_seconds}s...")
        time.sleep(self.inputs.retry_delay_seconds)
        return True

    def _finalize_workflow_execution(self) -> None:
        current = self.current_execution

        if current.metadata.end_time is None:
            current.metadata.end_time = datetime.now()

        if current.status == StatusCodes.IN_PROGRESS:
            current.status = StatusCodes.COMPLETED

    def _finalize_workflow(self) -> None:
        self._log_workflow_summary()

        if self.inputs.auto_generate_md:
            generate_workflow_documentation(
                workflow=self,
                md_output_path=self.inputs.md_file_path,
                diagram_output_path=self.inputs.diagram_file_path,
            )

    def _iterate_nodes(self) -> None:
        all_nodes: list[Node] = list(self._graph.nodes.values())
        running_threads: dict[str, threading.Thread] = {}
        scheduled: set[str] = set()
        completed: set[Node] = set()
        node_errors: dict[str, Exception] = {}

        while True:
            runnable = [
                n for n in all_nodes
                if n.name not in scheduled and self._graph.can_node_run(n, completed)
            ]

            for node in runnable:
                children = self._graph.get_children(node)
                if any(child.name in running_threads for child in children):
                    raise UnsupportedGraphScenario("Attempted to start a node while one of its children is running")

                def _target(n: Node) -> None:
                    try:
                        self.run_node(n)
                    except Exception as e:  # noqa: BLE001 - we intentionally capture to decide at workflow level
                        node_errors[n.name] = e

                thread = threading.Thread(target=_target, args=(node,), daemon=True)
                running_threads[node.name] = thread
                scheduled.add(node.name)
                thread.start()

            for name, thread in list(running_threads.items()):
                if thread.is_alive():
                    continue

                thread.join()
                node = self._graph.nodes[name]
                completed.add(node)
                self.current_execution.output.node_to_executions[node.name] = node.executions
                self._log_node_summary(node)
                if node.last_execution.status != StatusCodes.COMPLETED:
                    if self._all_execution_groups_dead():
                        raise node_errors.get(name) or Exception(
                            str(node.last_execution.error) if node.last_execution.error else f"Node {name} failed"
                        )
                del running_threads[name]

            if not running_threads:
                pending_runnable = [
                    n for n in all_nodes
                    if n.name not in scheduled and self._graph.can_node_run(n, completed)
                ]
                if not pending_runnable:
                    break

    def run_node(self, node: Node) -> None:
        node._set_workflow_context(workflow_input=self.inputs, workflow_metadata=self.metadata)

        self._log_node_start(node)
        node.execute()

    def _log_workflow_start(self) -> None:
        if not self.inputs.verbose:
            return

        cmd_line_line = f"Command: {build_cli_command_from_workflow_input(self.inputs, endpoint_name=self.endpoint_name)}\n" if self.endpoint_type == EndpointType.CLI else ""
        run_id_line = f"Run ID: {self.run_id}\n" if self.endpoint_type == EndpointType.API else ""
        self._logger.info(f"{'-'*30}\n"
                          f"Executing - Workflow: {self.name}, version: {self.version}\n"
                          f"Endpoint Type: {self.endpoint_type}\n"
                          f"Endpoint Name: {self.endpoint_name}\n"
                          f"{cmd_line_line}"
                          f"{run_id_line}"
                          f"Inputs: {self.inputs}\n"
                          f"{'-'*30}")

    def _log_workflow_summary(self) -> None:
        if not self.inputs.verbose:
            return

        latest = self.last_execution
        prev_status_lines = [
            f"  - Attempt #{i}: status={ex.status} duration={ex.metadata.process_time}"
            for i, ex in enumerate(self._executions[:-1], 1)
        ] if self._executions and len(self._executions) > 1 else []

        prev_section = f"Previous executions:\n{chr(10).join(prev_status_lines)}\n" if prev_status_lines else ""

        self._logger.info(f"{'-'*30}\n"
                          f"Summary - Workflow: {self.name} - Attempt #{self.attempt}\n"
                          f"Execution: {latest}\n"
                          f"{prev_section}"
                          f"{'-'*30}")

    def _log_node_start(self, node: Node) -> None:
        if not self.inputs.verbose:
            return

        self._logger.info(f"{'-'*30}\n"
                          f"Executing - Node: {node.name}\n"
                          f"Node Configuration:\n"
                          f"Timeout Seconds: {node.timeout_seconds}\n"
                          f"Retries: {node.max_retries}, Retry Delay Seconds: {node.retry_delay_seconds}\n"
                          f"{'-'*30}")

    def _log_node_summary(self, node: Node) -> None:
        latest = node.last_execution if node.last_execution else None
        if not self.inputs.verbose:
            status = latest.status if latest else StatusCodes.UNKNOWN
            duration = latest.metadata.process_time if latest else None
            if status == StatusCodes.COMPLETED:
                self._logger.info(f"Node {node.name} completed in {duration}")
            else:
                err_cls = latest.error.exception_class_name if latest and latest.error else None
                err_msg = latest.error.exception_message if latest and latest.error else None
                self._logger.info(
                    f"Node {node.name} failed ({status.name}) in {duration} error={err_cls}: {err_msg}"
                )
            return

        prev_failures = [
            f"\t\t- Attempt #{i}: status={ex.status} duration={ex.metadata.process_time} "
            f"error={ex.error.exception_class_name if ex.error else None}: "
            f"{ex.error.exception_message if ex.error else None}"
            for i, ex in enumerate(node.executions[:-1], 1)
            if ex.status != StatusCodes.COMPLETED
        ]
        prev_failures_str = '\n'.join(prev_failures)

        prev_section = (
            f"Previous execution failures:\n{prev_failures_str}\n"
            if prev_failures else ""
        )

        self._logger.info(
            f"{'-'*30}\n"
            f"Summary - Node: {node.name} - Execution #{node.attempt}\n"
            f"Execution: {latest}\n"
            f"{prev_section}"
            f"{'-'*30}"
        )

    def on_start(self) -> None:
        """Called when a workflow execution starts."""
        pass

    def on_success(self) -> None:
        """Called when a workflow execution completes successfully."""
        pass

    def on_failure(self, error: Exception) -> None:
        """Called when a workflow execution fails."""
        pass

    def on_finish(self) -> None:
        """Called at the end of any workflow execution (success or failure)."""
        pass
